import hashlib
import inspect
import os
import re
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

import databutton.helpers.parse as parse


def hash(file):
    BLOCK_SIZE = 65536  # The size of each read from the file

    file_hash = (
        hashlib.sha256()
    )  # Create the hash object, can use something other than `.sha256()` if you wish
    with open(file, "rb") as f:  # Open the file to read it's bytes
        fb = f.read(BLOCK_SIZE)  # Read from the file. Take in the amount declared above
        while len(fb) > 0:  # While there is still data being read from the file
            file_hash.update(fb)  # Update the hash
            fb = f.read(BLOCK_SIZE)  # Read the next block from the file

    return file_hash.hexdigest()


@dataclass
class DataframeArtifact:
    route: str
    name: str
    uid: str
    sourcefile: str
    importfile: str
    modulename: str
    last_updated_millis: int
    file_dependencies: List[str] = field(default_factory=lambda: [])
    file_version: str = ""


# Global var to store all dataframes
_dataframes: List[DataframeArtifact] = []


def dataframe(route: str, name: str = None):
    def app(func):
        df_name = name if name is not None else func.__name__
        codelines = inspect.getsourcelines(func)
        sourcefile = inspect.getsourcefile(func)
        imports = parse.get_library_dependencies_for_app(inspect.getsourcefile(func))
        # Make sure that the data folder exists
        Path(".databutton/data").mkdir(exist_ok=True, parents=True)
        filename = ".databutton/data/" + df_name + ".py"
        modulename = df_name

        artifact = DataframeArtifact(
            route=route,
            file_dependencies=[],
            sourcefile=sourcefile,
            importfile=filename,
            modulename=modulename,
            name=df_name,
            last_updated_millis=int(time.time() * 1000),
            uid=route.replace("/", "-").strip("-"),
        )

        for line in codelines[0]:
            for r in re.findall(r"\'.*?\'", line):
                r = r.replace("'", "")
                if os.path.isfile(r):
                    artifact.file_dependencies.append(r)
                    df_hash = hash(r)
                    artifact.file_version = df_hash
                    # print(line[:-2] + ' #file-version: ' + df_hash + '\n')  # TODO write this to the source-file

        with open(filename, "w+") as file:
            for line in imports:
                file.write(line)
            file.write("from functools import lru_cache \n")
            file.write("from fastapi import APIRouter \n\n")
            file.write("router = APIRouter()\n\n")
            file.write("@lru_cache(maxsize=None)\n")
            for line in codelines[0][1:]:
                file.write(line)

            file.write("\n\n")
            file.write("@router.get('" + route + "')\n")
            file.write("def loadThis():\n")
            file.write("    df = " + codelines[0][1][4:].replace(":", "") + "\n")
            file.write("    return df.to_json(orient='records')")

        _dataframes.append(artifact)

        init_files = ["./.databutton/data/__init__.py", "./.databutton/__init__.py"]
        for init_file in init_files:
            if os.path.exists(init_file) is False:
                open(init_file, "a").close()

    return app
